package ps4;

import java.awt.image.*;
import java.awt.*;

/**
 * Super class for all filter effects. Contains initialization, update, and
 * color access methods. Extends frame to make use of MediaTracker for loading.
 * 
 * @author Daniel Marcus
 */

public abstract class Filter {
	// OVERVIEW: A Filter represents an image and provides a technique for altering it.
	//     A Filter may be in one of three states: uninitialized, initialized,
	//     and applied.  An initialized or applied filter has an associated image;
	//     and a Pixels object that represents the pixel data (possibly modified 
	//     by the filter) in the image.
	
	private static final long serialVersionUID = 1L;

	private boolean initialized = false;
	protected String name;
	protected Image image; // Initial image to be filtered
	protected Pixels pixels; // Uncompressed pixel data

	// Rep Invariant
	//    initialized ==> image != null and pixels != null
	//    !initialized ==> image == null and pixels == null
	
	public Filter() 
	// EFFECTS: Initializes this to an uninitialized filter.
	{
		; 
	}

	final public void setImage(Image p_image, String p_name)
	// REQUIRES: this is uninitialized
	// MODIFIES: this
	// EFFECTS: Sets the image for this to p_image; sets this to the initialized state. 
	{
		assert !initialized;
		assert (p_image != null);
		
		name = p_name;
		image = p_image;
		pixels = new Pixels(image);
		initialized = true;
	}

	protected int getImageWidth() 
	// REQUIRES: this is initialized
	// EFFECTS: returns the width of the image of this
	{
		assert initialized;
		return image.getWidth(null);
	}

	protected int getImageHeight() 
	// REQUIRES: this is initialized;
	// EFFECTS: returns the height of the image of this
	{
		assert initialized;
		return image.getHeight(null);
	}

	protected abstract void filter();
      // REQUIRES: this must be initialized
	  // EFFECTS: alters the image in a manner specified by the filter.

	final public Image apply()
  	   // REQUIRES: this must be initialized with an image
	   // EFFECTS: returns the Image resulting from applying the filter to the image.
	{
		assert initialized;
		filter();
		// Need a component to create an image
		return new Container().createImage(new MemoryImageSource(pixels.getWidth(), pixels
				.getHeight(), pixels.convertTo1D(), 0, pixels.getWidth()));
	}

	public String getFilterName() {
		return "basic";
	}

	protected Color getColor(int row, int col)
	   // REQUIRES: row, col is a valid location in the image; this is initialized
	   // EFFECTS: Returns the color at location (row, col).
	{
		assert initialized;
		return pixels.getColor(row, col);
	}
	
	protected int getAlpha(int row, int col) 
    // REQUIRES: row, col is a valid location in the image; this is initialized
    // EFFECTS: Returns the alpha (transparency) value at location (row, col).
	{
		assert initialized;
		return pixels.getAlpha(row, col);
	}

	protected int getRed(int row, int col) 
	   // REQUIRES: row, col is a valid location in the image; this is initialized.
	   // EFFECTS: Returns the red value at location (row, col).
	{
		assert initialized;
		return pixels.getRed(row, col);
	}

	protected int getGreen(int row, int col) 
	   // REQUIRES: row, col is a valid location in the image; this is initialized.
	   // EFFECTS: Returns the green value at location (row, col).
	{
		assert initialized;
		return pixels.getGreen(row, col);
	}

	protected int getBlue(int row, int col) 
	   // REQUIRES: row, col is a valid location in the image; this is initialized
	   // EFFECTS: Returns the blue value at location (row, col).
	{
		assert initialized;
		return pixels.getBlue(row, col);
	}

	protected void setAlpha(int row, int col, int val)
	   // REQUIRES: row, col is a valid location in the image; 
	   //            0 <= val < 256; this is initialized 
	   // MODIFIES: this
	   // EFFECTS: Sets the alpha value at row, col to val
	{
		assert initialized;
		pixels.setAlpha(row, col, val);
	}

	protected void setColors(int row, int col, int r, int g, int b) 
	   // REQUIRES: row, col is a valid location in the image; 
	   //           0 <= r, g, b < 256; this is initialized 
	   // MODIFIES: this
	   // EFFECTS: Sets the color values at row, col to r, g, b
	{
		assert initialized;
		pixels.setColors(row, col, r, g, b);
	}
	
	protected void setColor(int row, int col, Color c)
	// REQUIRES: row, col is a valid location in the image; this is initialized
	// MODIFIES: this
	// EFFECTS: Sets the color at row, col to c
	{
		assert initialized;
		setColors(row, col, c.getRed(), c.getGreen(), c.getBlue());
	}

	public String getImageName() 
	// EFFECTS: Returns the image name associated with the filter.
	{
		if (name == null) {
			return "No Image";
		} else {
			return name;
		}
	}
}
