package ps4;

import java.awt.Component;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.Graphics2D;
import java.awt.Image;
import java.awt.MediaTracker;
import java.awt.Toolkit;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Vector;

import javax.imageio.ImageIO;
import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.DefaultListModel;
import javax.swing.ImageIcon;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JSplitPane;
import javax.swing.ListSelectionModel;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;

/**
 * @author drew
 * 
 */

public class GUI {
	// OVERVIEW: The GUI class manages the graphical user interface for
	// the application.

	static private final String[] effectNames = { "Blur", "Brighten", "Flip", "Greyscale",
			"Add", "Tile" };

	static private final String filterPackageName = "ps4";

	private Vector<String> effects;
	private JFrame rootFrame;
	private JMenuBar menuBar;
	private JPanel displayPane;
	private JList images;

	private DefaultListModel imagesModel;
	private ImageCellRenderer cr;

	// Main Panel Dimensions
	private static final int XSIZE = 800;
	private static final int YSIZE = 600;

	// Thumbnail Dimensions
	private static final int THUMBS_WIDTH = 160; // panel width
	private static final int THUMB_WIDTH = 48;
	private static final int THUMB_HEIGHT = 48;

	private GUIHandler handler;

	private String getFilterClass(String filter) 
	// EFFECTS: Returns the class name for filter.  Filters must
	//    follow the naming convention: ps4.<filter>Filter
	{
		return filterPackageName + "." + filter + "Filter";
	}
	public void reportError(String msg) {
		JOptionPane.showMessageDialog(rootFrame, msg, "Error",
				JOptionPane.WARNING_MESSAGE);
	}

	public void reportBug(String msg) {
		JOptionPane.showMessageDialog(rootFrame, msg, "Bug",
				JOptionPane.ERROR_MESSAGE);
	}

	public class GUIHandler implements ListSelectionListener, ActionListener {
		private JFrame rootFrame;
		private JList images;
		private DefaultListModel imagesModel;
		private ImageCellRenderer cr;
		private ArrayList<Image> workingImages;

		public GUIHandler(JFrame rootFrame, JList images,
				DefaultListModel imagesModel, ImageCellRenderer cr) {
			this.rootFrame = rootFrame;
			this.images = images;
			this.imagesModel = imagesModel;
			this.cr = cr;
			this.workingImages = new ArrayList<Image>();
		}
		
		private int [] selectMultipleImages() {
			JList simages = new JList(imagesModel);
			simages.setSelectionMode(ListSelectionModel.MULTIPLE_INTERVAL_SELECTION);
			simages.setCellRenderer(cr);
			JOptionPane.showMessageDialog(rootFrame, simages, "Select Images", JOptionPane.PLAIN_MESSAGE);
			return simages.getSelectedIndices();
		}

		private void displayImage(Image image, String name) {
			int imageWidth = image.getWidth(null);
			int imageHeight = image.getHeight(null);
			String label = name + " - " + imageWidth + " x " + imageHeight;
			
			int maxWidth = (int) (displayPane.getWidth() * 0.95);
			int maxHeight = (int) (displayPane.getHeight() * 0.95);

			if (imageWidth > maxWidth || imageHeight > maxHeight)
				image = Thumbnail.getScaledImage(image, maxWidth, maxHeight);

			displayPane.removeAll();
			displayPane.setLayout(new BoxLayout(displayPane, BoxLayout.PAGE_AXIS));

			displayPane.add(Box.createRigidArea(new Dimension(0, 8)));
			
			ImageIcon display = new ImageIcon(image);
			JLabel dlabel = new JLabel(display);
			dlabel.setAlignmentX(Component.CENTER_ALIGNMENT);
			displayPane.add(dlabel);
			displayPane.add(Box.createRigidArea(new Dimension(0, 5)));
			JLabel dtext = new JLabel(label);
			dtext.setAlignmentX(Component.CENTER_ALIGNMENT);
			displayPane.add(dtext);
			displayPane.validate();
			displayPane.updateUI();
		}

		private String getExtension(File file) {
			String filename = file.getName();
			int i = filename.lastIndexOf('.');
			if (i <= 0) {
				return null;
			} else {
				return filename.substring(i + 1).toLowerCase();
			}
		}

		private void addImage(Image image, String name) {
			Image thumbImage = Thumbnail.getScaledImage(image, THUMB_WIDTH,
					THUMB_HEIGHT);
			imagesModel.addElement(name);
			cr.addImage(thumbImage);
			workingImages.add(image);
			images.setSelectedIndex(workingImages.size() - 1);
			updateSelectedImage();
		}

		private void updateSelectedImage() {
			int idx = images.getSelectedIndex();
			if (idx >= 0) {
				Image image = workingImages.get(idx);
				displayImage(image, (String) imagesModel.get(idx));
			} else {
				displayPane.removeAll();
				displayPane.add(new JLabel("No Image Selected"));
				displayPane.updateUI();
			}
		}

		public void valueChanged(ListSelectionEvent event) {
			if (!event.getValueIsAdjusting()) {
				updateSelectedImage();
			}
		}

		/*
		 * Menu controls
		 */
		public void actionPerformed(ActionEvent event) {
			String command = event.getActionCommand();

			if (command.equals("open")) {
				JFileChooser chooser = new JFileChooser();
				chooser.setMultiSelectionEnabled(true);

				int openStatus = chooser.showOpenDialog(rootFrame);
				if (openStatus != JFileChooser.APPROVE_OPTION)
					return;
				File[] chosenFiles = chooser.getSelectedFiles();
				for (int i = 0; i < chosenFiles.length; i++) {
					File file = chosenFiles[i];
					String filename = file.getAbsolutePath();
					Image image = Toolkit.getDefaultToolkit()
							.getImage(filename);

					// This is needed to wait for the image to load
					MediaTracker mediaTracker = new MediaTracker(
							new Container());
					mediaTracker.addImage(image, 0);
					try {
						mediaTracker.waitForID(0);
					} catch (InterruptedException e) {
						; // Ignore the exception (the image might not be
						// loaded)
					}
					if (image == null) {
						reportError("Unable to load image: " + file.getName());
					} else if (image.getWidth(null) <= 0
							|| image.getHeight(null) <= 0) {
						reportError("Empty image: " + file.getName());
					} else {
						addImage(image, file.getName());
					}
				}
			} else if (command.equals("save")) {
				// Remove selected image(s)
				int idx = images.getSelectedIndex();
				if (idx < 0) {
					reportError("An image must be selected to save.");
					return;
				}

				JFileChooser chooser = new JFileChooser();
				chooser.setMultiSelectionEnabled(true);

				int chooseStatus = chooser.showSaveDialog(rootFrame);
				if (chooseStatus != JFileChooser.APPROVE_OPTION) {
					return;
				}
				File selectedFile = chooser.getSelectedFile();

				String ext = getExtension(selectedFile);
				if (ext == null) {
				    ext = "jpg";
				    selectedFile = new File(selectedFile.getAbsolutePath() + ".jpg");
				}
				if (!ext.equals("jpg")) {
					reportError("Save file must be a .jpg file");
					return;
				}
				Image image = workingImages.get(idx);
				BufferedImage bi = new BufferedImage(image.getWidth(null),
						image.getHeight(null), BufferedImage.TYPE_INT_RGB);
				Graphics2D g2 = bi.createGraphics();
				g2.drawImage(image, 0, 0, null);
				g2.dispose();
				try {
					ImageIO.write(bi, "jpg", selectedFile);
				} catch (IOException ioe) {
					reportError("Error writing image: " + ioe.getMessage());
				}
			} else if (command.equals("remove")) {
				// Remove selected image(s)
				int idx = images.getSelectedIndex();
				if (idx < 0) {
					reportError("An image must be selected to apply an effect.");
					return;
				}
				workingImages.remove(idx);
				images.clearSelection();
				updateSelectedImage();
				imagesModel.remove(idx);
				cr.removeImage(idx);
			} else if (effects.contains(command)) {
				Class fclass;

				try {
					// We need the full class name including the package
					fclass = Class.forName(getFilterClass(command));
				} catch (ClassNotFoundException e) {
					reportBug("Cannot find filter class: " + command);
					return;
				} catch (ExceptionInInitializerError e) {
					reportBug("Error initializing class " + command + ": " + e);
					return;
				} catch (LinkageError e) {
					reportBug("Linkage error: " + command);
					return;
				}

				Filter f;

				try {
					f = (Filter) fclass.newInstance();
				} catch (ClassCastException e) {
					reportBug("Filter class is not a subtype of BasicFilter");
					return;
				} catch (InstantiationException e) {
					reportBug("Cannot create filter object: " + e);
					return;
				} catch (IllegalAccessException e) {
					reportBug("Illegal access creating filter object: " + e);
					return;
				}

				int idx = images.getSelectedIndex();
				if (idx < 0) {
					reportError("An image must be selected to apply an effect.");
					return;
				}
				f.setImage(workingImages.get(idx), (String) imagesModel
						.get(idx));

				if (f instanceof MultiFilter) {
					MultiFilter mf = (MultiFilter) f;
					int [] simages = selectMultipleImages();
					if (simages.length == 0) {
						reportError("No images selected");
						return;
					}
					for (int sindex : simages) {
						mf.addImage(workingImages.get(sindex));
					}
				}

				Image result = f.apply();
				if (result == null) {
					reportError("Error applying filter");
				} else {
					addImage(result, f.getImageName() + "/" + f.getFilterName());
				}
			} else if (command.equals("exit")) {
				System.exit(0);
			}
		}
	}

	public GUI() {
		// Initialize the effects vector to the provided effects
		effects = new Vector<String>();

		for (String effect : effectNames) {
			effects.add(effect);
		}

		rootFrame = new JFrame("cs205 Rhocasa");
		menuBar = new JMenuBar();

		imagesModel = new DefaultListModel();
		images = new JList(imagesModel);
		cr = new ImageCellRenderer();
		images.setCellRenderer(cr);
		images.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
		handler = new GUIHandler(rootFrame, images, imagesModel, cr);
		images.addListSelectionListener(handler);
		JScrollPane thumbsListPane = new JScrollPane(images);
		thumbsListPane.setPreferredSize(new Dimension(THUMBS_WIDTH, YSIZE));
		displayPane = new JPanel();
		JSplitPane mainPane = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT,
				thumbsListPane, displayPane);

		JMenuItem item;
		JMenu menu = new JMenu("File");
		menu.setMnemonic(KeyEvent.VK_F);
		item = new JMenuItem("Open Image");
		item.setMnemonic(KeyEvent.VK_O);
		item.setActionCommand("open");
		item.addActionListener(handler);
		menu.add(item);

		item = new JMenuItem("Save Image");
		item.setMnemonic(KeyEvent.VK_S);
		item.setActionCommand("save");
		item.addActionListener(handler);
		menu.add(item);

		item = new JMenuItem("Remove Image");
		item.setMnemonic(KeyEvent.VK_R);
		item.setActionCommand("remove");
		item.addActionListener(handler);
		menu.add(item);

		item = new JMenuItem("Exit");
		item.setActionCommand("exit");
		item.setMnemonic(KeyEvent.VK_X);
		item.addActionListener(handler);
		menu.add(item);
		menuBar.add(menu);

		menu = new JMenu("Effects");
		menu.setMnemonic(KeyEvent.VK_E);
		for (String effect : effects) {
			item = new JMenuItem(effect);
			item.setActionCommand(effect);
			item.addActionListener(handler);
			menu.add(item);
		}
		menuBar.add(menu);

		rootFrame.getContentPane().add(mainPane);
		rootFrame.setJMenuBar(menuBar);
		rootFrame.setSize(XSIZE, YSIZE);
		rootFrame.setVisible(true);
	}

	public static void main(String[] args) {
		new GUI();
	}
}
