package ps4;

import java.awt.Color;
import java.awt.Image;
import java.awt.image.PixelGrabber;

/**
 * The Pixels class is used to convert to/from and hold uncompressed pixel data.
 * Holds seperate alpha, red, green, and blue values.
 * @author Daniel Marcus
 */

public class Pixels {
	// OVERVIEW: A Pixels object represents the pixel data for an image.  
	//    A typical Pixels is
	//       [ [ < R_0_0, G_0_0, B_0_0, A_0_0 >, < R_1_0, G_1_0, B_1_0, A_1_0 >, 
	//           ..., < R_width-1_0, G_width-1_0, B_width-1_0, A_width-1_0 > ],
	//         [ < R_0_1, G_0_1, B_0_1, A_0_1 >, < R_1_1, G_1_1, B_1_1, A_1_1 >, 
	//           ..., < R_width-1_1, G_width-1_1, B_width-1_1, A_width-1_1 > ],
	//         ...
	//         [ < R_0_height-1, G_0_height-1, B_0_height-1, A_0_height-1 >, ... ]]
	//   The value R_x_y is the red intensity at location x, y.
	//             G_x_y is the green intensity at location x, y.
	//             B_x_y is the blue intensity at location x, y.
	//             A_x_y is the alpha value (transparency) at location x, y.
	
	// Rep:
	private int height;			// Number image rows
	private int width;			// Number image columns
	private short[][][] data;	// Uncompressed image data (byte in Java is signed, so awkward for this)
	
	// Rep Invariant:
	//    data.length = height
	//    data[i].length = width for all 0 <= i < height
	//    data[i][j].length = 4 for all 0 <= i < height, 0 <= j < width
	//    0 <= data[i][j][k] <= 255 for all elements of data (guaranteed by byte datatype)
	
	/* 
	 * Final array indecies to reduce confusion
	 * Alpha value is pixel transparency: 255 = fully opaque, 0 = fully transparent
	 * RGB values correspond to color strengths; 255 = max color, 0 = no color
	 */
	protected final int ALPHA = 0;
	protected final int RED   = 1;
	protected final int GREEN = 2;
	protected final int BLUE  = 3;

	private void initializePixelData(int [] p_pix)
	{
		for (int x = 0; x < height; x++) {
			for (int y = 0; y < width; y++) {			
				int index = (x * width) + y;
				data[x][y][ALPHA] = (short) ((p_pix[index] >> 24) & 0xff);
				data[x][y][RED]   = (short) ((p_pix[index] >> 16) & 0xff);
				data[x][y][GREEN] = (short) ((p_pix[index] >>  8) & 0xff);
				data[x][y][BLUE]  = (short) ((p_pix[index]      ) & 0xff);
			}
		}
	}
	
	public Pixels(int[] p_pix, int p_width, int p_height)
	// REQUIRES: p_width and p_height are positive
	//           p_pix.length = p_width * p_height
	// EFFECTS: Initializes this to pixels representing the
	//     image given by the compressed array p_pix with
	//     
	{
		assert p_width > 0;
		assert p_height > 0;
		assert p_pix.length == p_width * p_height;
		
		width = p_width;
		height = p_height;

		data = new short[height][width][4];
		initializePixelData(p_pix);
	}
	
	public Pixels(Pixels p)
	// EFFECTS: Initializes this to a copy of p
	{
		this.width = p.width;
		this.height = p.height;
		this.data = new short[height][width][4];
		for (int i = 0; i < height; i++) {
			for (int j = 0; j < width; j++) {
				for (int k = 0; k < 4; k++) {
					data[i][j][k] = p.data[i][j][k];
				}
			}
		}
	}
	
	public Pixels(Image image)
	// EFFECTS: Initializes this to the pixels in image
	{
		width = image.getWidth(null);
		height = image.getHeight(null);
		int [] pix = new int[width * height];
		PixelGrabber pg = new PixelGrabber(image, 0, 0, width, height,
				pix, 0, width);
		try {
			pg.grabPixels();
		} catch (InterruptedException e) {
			System.out.print("PixelGrabber interrupted.");
			System.exit(1);
		}
		data = new short[height][width][4];
		initializePixelData(pix);
	}
	
	public int getHeight() {
		return height;
	}
	
	public int getWidth() {
		return width;
	}
	
	/**
	 * Compresses pixel data from four seperate ints to one int
	 * @return : 1D int array with data compressed to image format
	 */
	public int[] convertTo1D()
	// EFFECTS: Returns a one-dimensional compressed integer representation 
	//     of the pixels suitable for use in a MemoryImageSource.
	{
		int [] res = new int[width*height];
		for (int row = 0; row < height; row++) {
			for (int col = 0; col < width; col++) {				
				res[row*width + col] = 
					(((int) (data[row][col][ALPHA] << 24)) & 0xff000000) |
					(((int) (data[row][col][RED]   << 16)) & 0x00ff0000) |
					(((int) (data[row][col][GREEN] <<  8)) & 0x0000ff00) |
					(((int) (data[row][col][BLUE]       )) & 0x000000ff);
			}
		}
		
		return res;
	}
		
	/**
	 * Accessor
	 * @param x : row index
	 * @param y : column index
	 * @return  : alpha value at index
	 */
	public int getAlpha(int x, int y)
	{
		return data[x][y][ALPHA];
	}
	
	/**
	 * Accessor
	 * @param x : row index
	 * @param y : column index
	 * @return  : red value at index
	 */
	public int getRed(int x, int y)
	{
		return data[x][y][RED];
	}
	
	/**
	 * Accessor
	 * @param x : row index
	 * @param y : column index
	 * @return  : green value at index
	 */
	public int getGreen(int x, int y)
	{
		return data[x][y][GREEN];
	}
	
	/**
	 * Accessor
	 * @param x : row index
	 * @param y : column index
	 * @return  : blue value at index
	 */
	public int getBlue(int x, int y)
	{
		return data[x][y][BLUE];
	}

	private void setDataValue(int row, int col, int component, int val)
	   // REQUIRES: 0 <= val <= 255 and (row, col) is a valid pixel location
	   //	           and component is one of ALPHA, RED, BLUE, or GREEN
	   // MODIFIES: this
	   // EFFECTS: Sets the component value of the pixel at (row, col) in this to val.
	{
		assert row >= 0 && row < height;
		assert col >= 0 && col < width;
		assert val >= 0 && val < 256;
		data[row][col][component] = (short) val;
	}
	
	public void setAlpha(int row, int col, int val)
  	   // REQUIRES: 0 <= val <= 255 and (row, col) is a valid pixel location
	   // MODIFIES: this
	   // EFFECTS: Sets the alpha value of the pixel at (row, col) in this to val.
	{
		setDataValue (row, col, ALPHA, val);
	}
	
	public void setRed(int row, int col, int val)
	   // REQUIRES: 0 <= val <= 255 and (row, col) is a valid pixel location
	   // MODIFIES: this
	   // EFFECTS: Sets the red value of the pixel at (row, col) in this to val.
	{
		setDataValue (row, col, RED, val);
	}
	
	public void setGreen(int row, int col, int val)
	   // REQUIRES: 0 <= val <= 255 and (row, col) is a valid pixel location
	   // MODIFIES: this
	   // EFFECTS: Sets the green value of the pixel at (row, col) in this to val.
	{
		setDataValue (row, col, GREEN, val);
	}
	
	public void setBlue(int row, int col, int val)
	   // REQUIRES: 0 <= val <= 255 and (row, col) is a valid pixel location
	   // MODIFIES: this
	   // EFFECTS: Sets the blue value of the pixel at (row, col) in this to val.
	{
		setDataValue (row, col, BLUE, val);
	}
	
	public void setColors(int row, int col, int r, int g, int b)
	   // REQUIRES: 0 <= val <= 255 and (row, col) is a valid pixel location
	   // MODIFIES: this
	   // EFFECTS: Sets the color values of the pixel at (row, col) in this to r, g, b.
	{
		setRed(row, col, r);
		setGreen(row, col, g);
		setBlue(row, col, b);
	}

	public Color getColor(int row, int col) {
		return new Color (getRed (row, col), getGreen (row, col), getBlue (row, col));
	}

	public void setColor(int row, int col, Color color) {
		setColors (row, col, color.getRed(), color.getGreen(), color.getBlue());
	}
}
